/*
   Watch5 ARM7 Code
   Maarten Pennings - 2008 jan 2
   From Watch4, but changes to sound interfacing:
   - VblankHandler maps request i to channel i (instead of finding a free channel)
   - VblankHandler only calls startSound when snd->data[i].len>0.
   Signalling to Arm9
   - IPC->aux     is now the collection of BUSY flags.
   - IPC->battery is an alternating bit teling whether Arm7 has caught up with Arm9 (executing its play request)
*/

#include <nds.h>
#include <stdlib.h>

/* Protocol (Communication & Synchronisation) with Arm9 on sound
 IPC->aux expresses which sound channels are busy, i.e. bit n equals SCHANNEL_ENABLE of SCHANNEL_CR(n)
 Note that IPC->aux is updated by VcountHandler, so it lags behind.
 Therefore, when manipulating SCHANNEL_CR, also IPC->aux is updated.
*/
#define ARM7_ALTBITS     (IPC->battery         )
#define ARM7_ALTBIT(n)   (IPC->battery & BIT(n))
#define ARM7_BUSYBITS    (IPC->aux             )
#define ARM7_BUSYBIT(n)  (IPC->aux     & BIT(n))

//---------------------------------------------------------------------------------
void startSound(int sampleRate, const void* data, u32 bytes, u8 channel, u8 vol,  u8 pan, u8 format) {
//---------------------------------------------------------------------------------
  SCHANNEL_TIMER(channel)  = SOUND_FREQ(sampleRate);
  SCHANNEL_SOURCE(channel) = (u32)data;
  SCHANNEL_LENGTH(channel) = bytes >> 2 ;
  SCHANNEL_CR(channel)     = SCHANNEL_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(vol) | SOUND_PAN(pan) | (format==1?SOUND_8BIT:SOUND_16BIT);
  // Arm9 signalling
  ARM7_BUSYBITS |= BIT(channel);  // Arm7 is busy on 'channel'
  ARM7_ALTBITS  ^= BIT(channel);  // Signal (alternating bit) that the Arm7 processed the 'Arm9 start sound command'
}


int vcount;
touchPosition first,tempPos;

//---------------------------------------------------------------------------------
void VcountHandler() {
//---------------------------------------------------------------------------------
  static int lastbut = -1;
  u32 i;
  uint8 ct[sizeof(IPC->time.curtime)];
  uint16 but=0, x=0, y=0, xpx=0, ypx=0, z1=0, z2=0, sndBusy;

  but = REG_KEYXY;
  rtcGetTimeAndDate((uint8 *)ct); // Reads 7 bytes, starting with 'hour' (not 'command')

  // Get the busy flag of all sound channels
  sndBusy=0;
  for( i=0; i<16; i++ )
    if( SCHANNEL_CR(i)&SCHANNEL_ENABLE ) sndBusy|=BIT(i);
  ARM7_BUSYBITS=sndBusy;

  if (!( (but ^ lastbut) & (1<<6))) {

    tempPos = touchReadXY();

    if ( tempPos.x == 0 || tempPos.y == 0 ) {
      but |= (1 <<6);
      lastbut = but;
    } else {
      x = tempPos.x;
      y = tempPos.y;
      xpx = tempPos.px;
      ypx = tempPos.py;
      z1 = tempPos.z1;
      z2 = tempPos.z2;
    }

  } else {
    lastbut = but;
    but |= (1 <<6);
  }

  if ( vcount == 80 ) {
    first = tempPos;
  } else {
    if (  abs( xpx - first.px) > 10 || abs( ypx - first.py) > 10 ||
        (but & ( 1<<6)) ) {

      but |= (1 <<6);
      lastbut = but;

    } else {
      IPC->mailBusy = 1;
      IPC->touchX     = x;
      IPC->touchY     = y;
      IPC->touchXpx   = xpx;
      IPC->touchYpx   = ypx;
      IPC->touchZ1    = z1;
      IPC->touchZ2    = z2;
      IPC->mailBusy = 0;
    }
  }
  IPC->buttons    = but;

  for(i=1; i<sizeof(ct); i++) {
    IPC->time.curtime[i] = ct[i-1];
  }

  vcount ^= (80 ^ 130);
  SetYtrigger(vcount);
}

//---------------------------------------------------------------------------------
void VblankHandler(void) {
//---------------------------------------------------------------------------------
  u32 i;
  TransferSound *snd = IPC->soundData;
  IPC->soundData = 0; // Flag (to our own next Vblank invocation) that we did startSound

  if( 0!=snd ) {
    if( snd->count>16 ) snd->count=16; // buffer overflow protection
    for( i=0; i<snd->count; i++ ) {
      s32 chan = i; // new style: TransferSound at index i is for sound channel i
      if( snd->data[i].len>0 ) {
        startSound(snd->data[i].rate, snd->data[i].data, snd->data[i].len, chan, snd->data[i].vol, snd->data[i].pan, snd->data[i].format);
      }
    }
  }
}

//---------------------------------------------------------------------------------
int main(int argc, char ** argv) {
//---------------------------------------------------------------------------------

  // Arm9 signalling
  ARM7_BUSYBITS = 0; // Arm7 is busy with none of the channels
  ARM7_ALTBITS  = 0; // Arbitrary 'alternating bit' settings for all channels (all 0) -- must be the same on Arm9

  // Reset the clock if needed
  rtcReset();

  //enable sound
  powerON(POWER_SOUND);
  SOUND_CR = SOUND_ENABLE | SOUND_VOL(0x7F);
  IPC->soundData = 0;

  irqInit();
  irqSet(IRQ_VBLANK, VblankHandler);
  SetYtrigger(80);
  vcount = 80;
  irqSet(IRQ_VCOUNT, VcountHandler);
  irqEnable(IRQ_VBLANK | IRQ_VCOUNT);

  // Keep the ARM7 idle
  while (1) swiWaitForVBlank();
}


